﻿using Android.Content;
using Android.Widget;
using Java.Util.Concurrent;
using Android.Graphics;
using CameraCharacteristics = Android.Hardware.Camera2.CameraCharacteristics;
using Android.Hardware.Camera2;
using Android.Media;
using Android.Views;
using Android.Util;
using Android.Hardware.Camera2.Params;
using Size = Android.Util.Size;
using Class = Java.Lang.Class;
using Rect = Android.Graphics.Rect;
using SizeF = Android.Util.SizeF;
using Android.Runtime;
using Android.OS;
using RectF = Android.Graphics.RectF;
using Android.Content.Res;
using Java.Nio;
using Image = Android.Media.Image;
using Microsoft.Maui.Controls;
using static Android.Graphics.ImageDecoder;
using Android.Hardware.Lights;
using static Android.Icu.Text.ListFormatter;
//using Byte = Java.Nio.ByteBuffer;
using Com.Google.Android.Exoplayer2.Core;
using static Java.Interop.JniEnvironment;
using Microsoft.Maui;
using OpenCV.Core;
using OpenCV.Android;

namespace AsCameraXMAUIDemo;

internal class MauiCameraView : GridLayout
{
    private readonly CameraView cameraView;
    private IExecutorService executorService;
    private bool started = false;
    private bool initiated = false;
    private readonly Context context;

    private readonly TextureView textureView;
    public CameraCaptureSession previewSession;
    private CaptureRequest.Builder previewBuilder;
    private CameraDevice cameraDevice;
    private readonly MyCameraStateCallback stateListener;
    private Size videoSize;
    private CameraManager cameraManager;
    private CameraCharacteristics camChars;
    private PreviewCaptureStateCallback sessionCallback;
    private readonly ImageAvailableListener photoListener;
    private HandlerThread backgroundThread;
    private Handler backgroundHandler;
    private ImageReader imgReader;


    public MauiCameraView(Context context, CameraView cameraView) : base(context)
    {
        OpenCVLoader.InitDebug();
        this.context = context;
        this.cameraView = cameraView;
        textureView = new(context);
        stateListener = new MyCameraStateCallback(this);
        photoListener = new ImageAvailableListener(this.cameraView);
        sessionCallback = new PreviewCaptureStateCallback(this);
        AddView(textureView);
        InitDevices();
    }

    private void InitDevices()
    {
        if (!initiated && cameraView != null)
        {
            cameraManager = (CameraManager)context.GetSystemService(Context.CameraService);

            foreach (var id in cameraManager.GetCameraIdList())
            {
                var cameraInfo = new CameraInfo { DeviceId = id, MinZoomFactor = 1 };
                var chars = cameraManager.GetCameraCharacteristics(id);
                if ((int)(chars.Get(CameraCharacteristics.LensFacing) as Java.Lang.Number) == (int)LensFacing.Back)
                {
                    cameraInfo.Name = "Back Camera";
                    cameraInfo.Position = CameraPosition.Back;
                    cameraInfo.MaxZoomFactor = (float)(chars.Get(CameraCharacteristics.ScalerAvailableMaxDigitalZoom) as Java.Lang.Number);
                    cameraInfo.HasFlashUnit = (bool)(chars.Get(CameraCharacteristics.FlashInfoAvailable) as Java.Lang.Boolean);
                    cameraInfo.AvailableResolutions = new();
                    try
                    {
                        int sensorOrientation = (int)(chars.Get(CameraCharacteristics.SensorOrientation) as Java.Lang.Integer);
                        cameraInfo.sensorOrientation = sensorOrientation;
                        float[] maxFocus = (float[])chars.Get(CameraCharacteristics.LensInfoAvailableFocalLengths);
                        SizeF size = (SizeF)chars.Get(CameraCharacteristics.SensorInfoPhysicalSize);
                        cameraInfo.HorizontalViewAngle = (float)(2 * Math.Atan(size.Width / (maxFocus[0] * 2)));
                        cameraInfo.VerticalViewAngle = (float)(2 * Math.Atan(size.Height / (maxFocus[0] * 2)));
                    }
                    catch
                    {
                    }
                    cameraInfo.AvailableResolutions.Add(new(1280, 720));
                    cameraView.Camera = cameraInfo;
                    break;
                }

            }
            executorService = Executors.NewSingleThreadExecutor();
            initiated = true;

            cameraView.RefreshDevices();
        }
    }

    private void StartPreview()
    {
        while (textureView.SurfaceTexture == null) Thread.Sleep(100);
        SurfaceTexture texture = textureView.SurfaceTexture;
        texture.SetDefaultBufferSize(videoSize.Width, videoSize.Height);
        previewBuilder = cameraDevice.CreateCaptureRequest(CameraTemplate.Preview);
        var surfaces = new List<OutputConfiguration>();
        var surfaces26 = new List<Surface>();
        var previewSurface = new Surface(texture);
        surfaces.Add(new OutputConfiguration(previewSurface));
        surfaces26.Add(previewSurface);
        previewBuilder.AddTarget(previewSurface);
        if (imgReader != null)
        {
            surfaces.Add(new OutputConfiguration(imgReader.Surface));
            surfaces26.Add(imgReader.Surface);
            previewBuilder.AddTarget(imgReader.Surface);

        }
        SessionConfiguration config = new((int)SessionType.Regular, surfaces, executorService, sessionCallback);
        cameraDevice.CreateCaptureSession(config);
    }
    private void UpdatePreview()
    {
        if (null == cameraDevice)
            return;

        try
        {
            previewBuilder.Set(CaptureRequest.ControlMode, Java.Lang.Integer.ValueOf((int)ControlMode.Auto));
            AdjustAspectRatio(videoSize.Width, videoSize.Height);
            ForceAutoFocus();
        }
        catch (CameraAccessException e)
        {
            e.PrintStackTrace();
        }
        cameraView.WhenCameraReady();
    }
    internal async Task<CameraResult> StartCameraAsync(Microsoft.Maui.Graphics.Size PhotosResolution)
    {
        var result = CameraResult.Success;
        if (initiated)
        {
            if (await CameraView.RequestPermissions())
            {
                if (started) StopCamera();
                if (cameraView.Camera != null)
                {
                    try
                    {
                        camChars = cameraManager.GetCameraCharacteristics(cameraView.Camera.DeviceId);

                        StreamConfigurationMap map = (StreamConfigurationMap)camChars.Get(CameraCharacteristics.ScalerStreamConfigurationMap);
                        videoSize = new Size(1280, 720); //ChooseVideoSize(map.GetOutputSizes(Class.FromType(typeof(ImageReader))));
                        var maxVideoSize = new Size(1280, 720); //ChooseMaxVideoSize(map.GetOutputSizes(Class.FromType(typeof(ImageReader))));
                        //if (PhotosResolution.Width != 0 && PhotosResolution.Height != 0)
                        //    maxVideoSize = new((int)PhotosResolution.Width, (int)PhotosResolution.Height);
                        imgReader = ImageReader.NewInstance(maxVideoSize.Width, maxVideoSize.Height, ImageFormatType.Yuv420888, 1);
                        backgroundThread = new HandlerThread("CameraBackground");
                        backgroundThread.Start();
                        backgroundHandler = new Handler(backgroundThread.Looper);
                        imgReader.SetOnImageAvailableListener(photoListener, backgroundHandler);

                        if (OperatingSystem.IsAndroidVersionAtLeast(28))
                            cameraManager.OpenCamera(cameraView.Camera.DeviceId, executorService, stateListener);
                        else
                            cameraManager.OpenCamera(cameraView.Camera.DeviceId, stateListener, null);

                        started = true;
                    }
                    catch
                    {
                        result = CameraResult.AccessError;
                    }
                }
                else
                    result = CameraResult.NoCameraSelected;
            }
            else
                result = CameraResult.AccessDenied;
        }
        else
            result = CameraResult.NotInitiated;

        return result;
    }

    internal CameraResult StopCamera()
    {
        CameraResult result = CameraResult.Success;
        if (initiated)
        {

            try
            {
                imgReader.Close();
                backgroundHandler.RemoveCallbacks(backgroundThread);
                backgroundThread?.QuitSafely();
                backgroundThread?.Join();
                backgroundThread = null;
                backgroundHandler = null;
                imgReader?.Dispose();
                imgReader = null;
            }
            catch { }
            try
            {
                previewSession?.StopRepeating();
                previewSession?.Dispose();
            }
            catch { }
            try
            {
                cameraDevice?.Close();
                cameraDevice?.Dispose();
            }
            catch { }
            previewSession = null;
            cameraDevice = null;
            previewBuilder = null;
            started = false;
        }
        else
            result = CameraResult.NotInitiated;
        return result;
    }
    internal void DisposeControl()
    {
        try
        {
            if (started) StopCamera();
            executorService?.Shutdown();
            executorService?.Dispose();
            RemoveAllViews();
            textureView?.Dispose();
            Dispose();
        }
        catch { }
    }

    public void UpdateMirroredImage()
    {
        if (cameraView != null && textureView != null)
        {
            if (cameraView.MirroredImage)
                textureView.ScaleX = -1;
            else
                textureView.ScaleX = 1;
        }
    } 
    internal void UpdateFlashMode()
    {
        if (previewSession != null && previewBuilder != null && cameraView.Camera != null && cameraView != null)
        {
            try
            {
                if (cameraView.Camera.HasFlashUnit)
                {
                    switch (cameraView.FlashMode)
                    {
                        case FlashMode.Auto:
                            previewBuilder.Set(CaptureRequest.ControlAeMode, (int)ControlAEMode.OnAutoFlash);
                            previewSession.SetRepeatingRequest(previewBuilder.Build(), null, null);
                            break;
                        case FlashMode.Enabled:
                            previewBuilder.Set(CaptureRequest.ControlAeMode, (int)ControlAEMode.On);
                            previewSession.SetRepeatingRequest(previewBuilder.Build(), null, null);
                            break;
                        case FlashMode.Disabled:
                            previewBuilder.Set(CaptureRequest.ControlAeMode, (int)ControlAEMode.Off);
                            previewSession.SetRepeatingRequest(previewBuilder.Build(), null, null);
                            break;
                    }
                }
            }
            catch (System.Exception)
            {
            }
        }
    }
    internal void ForceAutoFocus()
    {
        if (previewSession != null && previewBuilder != null && cameraView.Camera != null)
        {
            previewBuilder.Set(CaptureRequest.ControlAfMode, Java.Lang.Integer.ValueOf((int)ControlAFMode.Off));
            previewBuilder.Set(CaptureRequest.ControlAfTrigger, Java.Lang.Integer.ValueOf((int)ControlAFTrigger.Cancel));
            previewSession.SetRepeatingRequest(previewBuilder.Build(), null, null);
            previewBuilder.Set(CaptureRequest.ControlAfMode, Java.Lang.Integer.ValueOf((int)ControlAFMode.Auto));
            previewBuilder.Set(CaptureRequest.ControlAfTrigger, Java.Lang.Integer.ValueOf((int)ControlAFTrigger.Start));
            previewSession.SetRepeatingRequest(previewBuilder.Build(), null, null);

        }
    }

    private void AdjustAspectRatio(int videoWidth, int videoHeight)
    {
        Matrix txform = new();
        RectF viewRect = new(0, 0, Width, Height);
        float centerX = viewRect.CenterX();
        float centerY = viewRect.CenterY();
        RectF bufferRect = new(0, 0, videoHeight, videoWidth);
        bufferRect.Offset(centerX - bufferRect.CenterX(), centerY - bufferRect.CenterY());
        txform.SetRectToRect(viewRect, bufferRect, Matrix.ScaleToFit.Fill);
        float scale = (float)Width / videoHeight;
        AsCameraXManager.sharedInstance().viewWidth = Width;
        AsCameraXManager.sharedInstance().viewHeight = Height;
        AsCameraXManager.sharedInstance().viewScale = scale;
        txform.PostScale(scale, scale, centerX, centerY);
        textureView.SetTransform(txform);
    }

    protected override async void OnConfigurationChanged(Configuration newConfig)
    {
        base.OnConfigurationChanged(newConfig);
        if (started)
            await StartCameraAsync(cameraView.PhotosResolution);
    }
    private class MyCameraStateCallback : CameraDevice.StateCallback
    {
        private readonly MauiCameraView cameraView;
        public MyCameraStateCallback(MauiCameraView camView)
        {
            cameraView = camView;
        }
        public override void OnOpened(CameraDevice camera)
        {
            if (camera != null)
            {
                cameraView.cameraDevice = camera;
                cameraView.StartPreview();
            }
        }

        public override void OnDisconnected(CameraDevice camera)
        {
            camera.Close();
            cameraView.cameraDevice = null;
        }

        public override void OnError(CameraDevice camera, CameraError error)
        {
            camera?.Close();
            cameraView.cameraDevice = null;
        }
    }
    private class PreviewCaptureStateCallback : CameraCaptureSession.StateCallback
    {
        private readonly MauiCameraView cameraView;
        public PreviewCaptureStateCallback(MauiCameraView camView)
        {
            cameraView = camView;
        }
        public override void OnConfigured(CameraCaptureSession session)
        {
            cameraView.previewSession = session;
            cameraView.UpdatePreview();
        }
        public override void OnConfigureFailed(CameraCaptureSession session)
        {
        }
    }

    private class DecodeThreed : Java.Lang.Thread
    {
        private byte[] currentDecodeByte;
        private readonly CameraView cameraView;

        public DecodeThreed(CameraView camView)
        {
            cameraView = camView;
        }
        public override void Run()
        {
            base.Run();

            while (true)
            {
                if (currentDecodeByte != null)
                {
                    Log("输出");
                    cameraView.WhenOnImageAvailableReceivedByte(currentDecodeByte, 1280, 720);
                    currentDecodeByte = null;
                }
            }
        }
        public void Log(string message)
        {
            string timeStamp = DateTime.Now.ToString("yyyy-MM-dd HH:mm:ss fff");
            string logEntry = $"{timeStamp} - {message}";
            Console.WriteLine(logEntry);
        }
        private static int mWidth = 0;
        private static int mHeight = 0;
        private static int mBytes = 0;
        private static byte[] mData = null;
        static byte[] GetBuffer(int width, int height, int nbytes)
        {
            if (width != mWidth || height != mHeight || mBytes != nbytes || mData == null)
            {
                mData = null;
                mWidth = width;
                mHeight = height;
                mBytes = nbytes;
                mData = new byte[nbytes];
            }
            return mData;
        }
        public void setImage(Image image)
        {
            if (currentDecodeByte == null)
            {
                Log("赋值 解析");
                Image.Plane plane = image.GetPlanes()[0];
                ByteBuffer buffer = plane.Buffer;
                int y_plane_step = plane.RowStride; 
                Mat mGray = new Mat(image.Height, image.Width ,  CvType.Cv8uc1, buffer, y_plane_step);
                buffer.Clear();
                byte[] pData = new byte[(int)mGray.Total()];
                mGray.Get(0, 0, pData);
                mGray.Release();
                mGray = null;
                currentDecodeByte = pData;
                Log("赋值 解析完成");
            }
            else
            {
                Log("不解析");
            }

        }

    }
    class ImageAvailableListener : Java.Lang.Object, ImageReader.IOnImageAvailableListener
    {
        private readonly CameraView cameraView;
        DecodeThreed decodeThreed;
        public ImageAvailableListener(CameraView camView)
        {
            cameraView = camView;
            decodeThreed = new DecodeThreed(camView);
            decodeThreed.Start();

        }
        public void OnImageAvailable(ImageReader reader)
        {
            try
            {
                var image = reader?.AcquireNextImage();
                if (decodeThreed != null)
                {
                    decodeThreed.setImage(image);
                }
                image.Close();
            }
            catch
            {
            }
        }
    }
}


